@@ -37,7 +37,6 @@
 //! * [`impl Trait`](#impl-trait)
 //! * [`Mocking structs`](#mocking-structs)
 //! * [`Generic methods`](#generic-methods)
-//! * [`Methods with generic lifetimes`](#methods-with-generic-lifetimes)
 //! * [`Generic traits and structs`](#generic-traits-and-structs)
 //! * [`Associated types`](#associated-types)
 //! * [`Multiple and inherited traits`](#multiple-and-inherited-traits)
@@ -59,7 +58,7 @@
 //!     fn foo(&self, x: u32) -> u32;
 //! }
 //!
-//! fn call_with_four(x: &MyTrait) -> u32 {
+//! fn call_with_four(x: &dyn MyTrait) -> u32 {
 //!     x.foo(4)
 //! }
 //!
@@ -673,12 +672,15 @@
 //!
 //! ## Generic methods
 //!
-//! Generic methods can be mocked, too.  Effectively each generic method is an
-//! infinite set of regular methods, and each of those works just like any other
-//! regular method.  The expect_* method is generic, too, and usually must be
-//! called with a turbofish.  The only restrictions on mocking generic methods
-//! are that all generic parameters must be `'static`, and generic lifetime
-//! parameters are not allowed.
+//! Mocking generic methods is possible, but the exact process depends on
+//! whether the parameters are `'static`, non-`'static`, or lifetimes.
+//!
+//! ### With static parameters
+//!
+//! With fully `'static` parameters, the mock method is generic and so is its
+//! expect_* method.  The expect_* method usually must be called with a
+//! turbofish.  Expectations set with different generic parameters operate
+//! completely independently of one another.
 //!
 //! ```
 //! # use mockall::*;
@@ -697,7 +699,15 @@
 //! assert_eq!(-5, mock.foo(5i8));
 //! ```
 //!
-//! ## Methods with generic lifetimes
+//! ### With non-`static` type parameters
+//!
+//! Mocking methods with non-`'static` type parameters is harder.  The way
+//! Mockall does it is by turning the generic parameters into trait objects
+//! before evaluating expectations.  This makes the expect_* method concrete,
+//! rather than generic.  It also comes with many restrictions.  See
+//! [`#[concretize]`](attr.concretize.html) for more details.
+//!
+//! ### With generic lifetimes
 //!
 //! A method with a lifetime parameter is technically a generic method, but
 //! Mockall treats it like a non-generic method that must work for all possible
@@ -1136,7 +1146,7 @@
 use downcast::*;
 use std::{
     any,
-    fmt::{self, Debug, Formatter},
+    fmt::Debug,
     marker::PhantomData,
     ops::{Range, RangeFrom, RangeFull, RangeInclusive, RangeTo,
           RangeToInclusive},
@@ -1235,14 +1245,29 @@ pub mod examples;
 /// }
 /// ```
 ///
-/// Finally, `#[automock]` can also mock foreign functions.  This requires
-/// another metaitem to specify the mock module name.
+/// It can mock a module full of functions.  In this case, the mock functions
+/// will be found in a module whose name is prepended with `mock_`.
 ///
 /// ```
 /// # use mockall_derive::*;
-/// #[automock(mod mock_ffi;)]
-/// extern "C" {
-///     pub fn foo() -> u32;
+/// #[automock]
+/// mod mymod {
+///     pub fn foo() -> u32 {
+///        // ...
+///        # unimplemented!()
+///     }
+/// }
+/// ```
+/// Finally, `#[automock]` can also mock foreign functions.  This works just
+/// like mocking a module.
+///
+/// ```
+/// # use mockall_derive::*;
+/// #[automock]
+/// mod ffi {
+///     extern "C" {
+///         pub fn foo() -> u32;
+///     }
 /// }
 /// ```
 ///
@@ -1260,6 +1285,60 @@ pub mod examples;
 ///   to choose your own name for the mock structure.
 pub use mockall_derive::automock;
 
+/// Decorates a method or function to tell Mockall to treat its generic arguments
+/// as trait objects when creating expectations.
+///
+/// This allows users to use non-`'static` generic parameters, which otherwise
+/// can't be mocked.  The downsides of using this attribute are:
+///
+/// * Mockall can't tell if a parameter isn't `'static`, so you must annotate
+/// such methods with the `#[mockall::concretize]` attribute.
+/// * Generic methods will share expectations for all argument types.  That is,
+///   you won't be able to do `my_mock.expect_foo::<i32>(...)`.
+/// * It can't be used on methods with a closure argument (though this may be
+/// fixable).
+/// * Concretized methods' expectations may only be matched with `.withf` or
+/// `.withf_st`, not `.with`.
+/// * It only works for parameters that can be turned into a trait object.
+/// (may be fixable).
+/// * Mockall needs to know how to turn the function argument into a trait
+/// object.  Given a generic parameter `T`, currently supported patterns are:
+///   - `T`
+///   - `&T`
+///   - `&mut T`
+///   - `&[T]`
+///
+/// # Examples
+/// ```
+/// # use std::path::Path;
+/// # use mockall::{automock, concretize};
+/// #[automock]
+/// trait Foo {
+///     #[mockall::concretize]
+///     fn foo<P: AsRef<Path>>(&self, p: P);
+/// }
+///
+/// # fn main() {
+/// let mut mock = MockFoo::new();
+/// mock.expect_foo()
+///     .withf(|p| p.as_ref() == Path::new("/tmp"))
+///     .return_const(());
+/// mock.foo(Path::new("/tmp"));
+/// # }
+/// ```
+///
+/// NB: This attribute must be imported with its canonical name.  It won't work
+/// otherwise!
+/// ```compile_fail
+/// use mockall::concretize as something_else;
+/// #[mockall::automock]
+/// trait Foo {
+///     #[something_else]
+///     fn foo<T>(&self, t: T);
+/// }
+/// ```
+pub use mockall_derive::concretize;
+
 /// Manually mock a structure.
 ///
 /// Sometimes `automock` can't be used.  In those cases you can use `mock!`,
@@ -1311,8 +1390,8 @@ pub use mockall_derive::automock;
 /// ```
 ///
 /// When mocking a generic struct's implementation of a generic trait, use the
-/// same namespace for their generic parameters.  For example, if you wanted to
-/// mock `Rc`, do
+/// same name for their generic parameters.  For example, if you wanted to mock
+/// `Rc`, do
 /// ```
 /// # use mockall_derive::mock;
 /// mock!{
@@ -1398,28 +1477,38 @@ pub struct DefaultReturner<O>(PhantomData<O>);
     }
 }
 
+// Wrapper type to allow for better expectation messages for any type.
+// Will first try Debug, otherwise will print '?'
 #[doc(hidden)]
-pub struct MaybeDebugger<'a, T>(pub &'a T);
-::cfg_if::cfg_if! {
-    if #[cfg(feature = "nightly")] {
-        impl<'a, T> Debug for MaybeDebugger<'a, T> {
-            default fn fmt(&self, f: &mut Formatter<'_>)
-                -> Result<(), fmt::Error>
-            {
-                write!(f, "?")
-            }
-        }
-        impl<'a, T: Debug> Debug for MaybeDebugger<'a, T> {
-            fn fmt(&self, f: &mut Formatter<'_>) -> Result<(), fmt::Error> {
-                self.0.fmt(f)
-            }
-        }
-    } else {
-        impl<'a, T> Debug for MaybeDebugger<'a, T> {
-            fn fmt(&self, f: &mut Formatter<'_>) -> Result<(), fmt::Error> {
-                write!(f, "?")
-            }
-        }
+pub struct ArgPrinter<'a, T>(pub &'a T);
+
+#[doc(hidden)]
+pub struct DebugPrint<'a, T: Debug>(pub &'a T);
+impl<'a, T> Debug for DebugPrint<'a, T> where T: Debug {
+    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
+        Debug::fmt(self.0, f)
+    }
+}
+#[doc(hidden)]
+pub trait ViaDebug<T> where T: Debug { fn debug_string(&self) -> DebugPrint<'_, T>; }
+impl<'a, T: Debug> ViaDebug<T> for &ArgPrinter<'a, T> {
+    fn debug_string(&self) -> DebugPrint<'a, T> {
+        DebugPrint(self.0)
+    }
+}
+
+#[doc(hidden)]
+pub struct NothingPrint;
+impl Debug for NothingPrint {
+    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
+        write!(f, "?")
+    }
+}
+#[doc(hidden)]
+pub trait ViaNothing { fn debug_string(&self) -> NothingPrint; }
+impl<'a, T> ViaNothing for ArgPrinter<'a, T> {
+    fn debug_string(&self) -> NothingPrint {
+        NothingPrint
     }
 }
 
@@ -1625,7 +1714,7 @@ impl SeqInner {
     /// Verify that the call identified by `seq` was called in the correct order
     fn verify(&self, seq: usize, desc: &str) {
         assert_eq!(seq, self.satisfaction_level.load(Ordering::Relaxed),
-            "{}: Method sequence violation", desc)
+            "{desc}: Method sequence violation")
     }
 }
 
